require "date"

class DayTime
  include Enumerable
  include Comparable

  SECONDS_PER_DAY = 86_399

  HM_REGEXP  = /^(2[0-3]|[01]?[0-9]):([0-5]?[0-9])$/i
  HMS_REGEXP = /^(2[0-3]|[01]?[0-9]):([0-5]?[0-9]):([0-5]?[0-9])$/i

  attr_reader :seconds_since_midnight

  def initialize(arg)
    @seconds_since_midnight = case arg
                              when self.class
                                arg.seconds_since_midnight
                              when Time
                                seconds_from_time(arg)
                              when DateTime
                                seconds_from_time(arg.to_time)
                              when Integer
                                seconds_from_integer(arg)
                              when String
                                seconds_from_time_string(arg)
                              when Hash
                                seconds_from_hash(arg)
                              else
                                raise ArgumentError
                              end
  end

  def hours
    @seconds_since_midnight / (60 ** 2)
  end

  def minutes
    (@seconds_since_midnight % 60 ** 2) / 60
  end

  def seconds
    @seconds_since_midnight % 60
  end

  def <=>(other)
    seconds_since_midnight <=> other.seconds_since_midnight
  end

  def succ
    self + 1
  end

  def each
    1.upto(Float::INFINITY) { |i| yield self + i }
  end

  def +(obj)
    other = DayTime.new(obj)
    DayTime.new(@seconds_since_midnight + other.seconds_since_midnight)
  end

  def -(obj)
    other = DayTime.new(obj)
    DayTime.new((@seconds_since_midnight - other.seconds_since_midnight).abs)
  end

  def to_s
    [hours, minutes, seconds].map { |int| int.to_s.rjust(2, "0") }.join(":")
  end

  def to_i
    @seconds_since_midnight
  end

  private

  def seconds_from_integer(seconds_since_midnight)
    if seconds_since_midnight > SECONDS_PER_DAY
      (seconds_since_midnight % SECONDS_PER_DAY) - 1
    else
      seconds_since_midnight
    end
  end

  def seconds_from_time_string(str)
    h, m, s = if match = str.match(HMS_REGEXP)
                match.captures.map(&:to_i)
              elsif match = str.match(HM_REGEXP)
                match.captures.map(&:to_i) << 0
              else
                raise ArgumentError
              end

    seconds_from_hash(h: h, m: m, s: s)
  end

  def seconds_from_hash(hash)
    [[:hours, :h], [:minutes, :m]].each do |keys|
      unless keys.any? { |key| hash.key?(key) }
        raise ArgumentError(
          "Either a `#{keys[0]}` or `#{keys[2]} key is required"
        )
      end
    end

    h = hash[:hours]   || hash[:h]
    m = hash[:minutes] || hash[:m]
    s = hash[:seconds] || hash[:s] || 0

    seconds_from_integer(h * 60 ** 2 + m * 60 + s)
  end

  def seconds_from_time(time)
    seconds_from_hash(
      h: time.hour,
      m: time.min,
      s: time.sec
    )
  end
end

def DayTime(*argv)
  DayTime.new(*argv)
end
